import compas
import compas_rhino
from compas_rhino.install import install
from compas_rhino.uninstall import uninstall
from compas_rhino.install_plugin import install_plugin
from compas_rhino.uninstall_plugin import uninstall_plugin
import argparse
import os
import json
import sys
import importlib


PLUGIN_NAME = "3GS"
PACKAGES = ['compas', 'compas_rhino', 'compas_3gs', 'compas_skeleton']


def is_editable(project_name):
    """Is distribution an editable install?"""
    for path_item in sys.path:
        egg_link = os.path.join(path_item, "%s.egg-link" % project_name)
        if os.path.isfile(egg_link):
            return True
    return False


if __name__ == '__main__':

    parser = argparse.ArgumentParser(
        description='compas_3gs Installation command-line utility.')

    parser.add_argument('--remove_plugins', action='store_true', help="remove all existing plugins")
    parser.add_argument('--remove_packages', action='store_true', help="remove all existing compas packages")
    parser.add_argument('--rhino_version', default='6.0', choices=['6.0', '7.0'], help="remove all existing compas packages")
    args = parser.parse_args()

    print("\n", "-" * 10, "Checking packages", "-" * 10)

    for p in PACKAGES:
        try:
            importlib.import_module(p)
        except ImportError:
            print(p, "ERROR: cannot be imported, make sure it is installed")
            raise
        else:
            print('   {} {}'.format(p.ljust(20), "OK"))

    is_dev = is_editable("compas-3gs")
    print("compas_3gs is editable install: ", is_dev)

    if args.remove_plugins:
        print("\n", "-" * 10, "Removing existing plugins", "-" * 10)
        python_plugins_path = compas_rhino._get_python_plugins_path(args.rhino_version)
        print("Plugin location: ", python_plugins_path)
        plugins = os.listdir(python_plugins_path)
        for p in plugins:
            uninstall_plugin(p, version=args.rhino_version)

    if args.remove_packages:
        print("\n", "-" * 10, "Removing existing packages", "-" * 10)
        uninstall()

    print("\n", "-" * 10, "Installing compas_3gs python plugin", "-" * 10)

    plugin_path = os.path.dirname(__file__)
    plugin_path = os.path.join(plugin_path, 'ui/Rhino/3GS')
    plugin_path = os.path.abspath(plugin_path)

    if os.path.exists(plugin_path):
        python_plugins_path = compas_rhino._get_python_plugins_path(args.rhino_version)
        print("Plugin path:", python_plugins_path)
        install_plugin(plugin_path, version=args.rhino_version)
    else:
        raise RuntimeError("%s does not exist" % plugin_path)

    print("\n", "-" * 10, "Installing COMPAS packages", "-" * 10)

    if not is_dev:
        os.environ["CONDA_PREFIX"] = ""
        os.environ["CONDA_DEFAULT_ENV"] = ""
        os.environ["CONDA_EXE"] = ""

    print("CONDA_PREFIX", os.environ["CONDA_PREFIX"])
    print("CONDA_DEFAULT_ENV", os.environ["CONDA_DEFAULT_ENV"])
    print("CONDA_EXE", os.environ["CONDA_EXE"])

    install(packages=PACKAGES, version=args.rhino_version)

    print("\n", "-" * 10, "Installation is successful", "-" * 10)

    print("\n", "-" * 10, "Registering Installation", "-" * 10)

    os.makedirs(compas.APPDATA, exist_ok=True)
    register_json_path = os.path.join(compas.APPDATA, "compas_plugins.json")

    if os.path.exists(register_json_path):
        register_json = json.load(open(register_json_path))
        if not isinstance(register_json["Plugins"], dict):
            register_json["Plugins"] = {}
    else:
        register_json = {"Plugins": {}, "Current": None}

    plugin_info = {
        "dev": is_dev,
        "path": plugin_path,
        "python": sys.executable,
        "packages": {},
    }

    for name in PACKAGES:
        package = importlib.import_module(name)
        plugin_info["packages"][name] = {"version": package.__version__}

    print(plugin_info)

    print("registering to", register_json_path)

    register_json["Plugins"][PLUGIN_NAME] = plugin_info
    register_json["Current"] = PLUGIN_NAME

    print(" " * 4, plugin_path, "is registered")

    for name in register_json["Plugins"]:
        plugin = register_json["Plugins"][name]
        if not os.path.exists(plugin["path"]):
            del register_json["Plugins"][name]
            print("    Removed un-existed path: ", plugin_path)

    json.dump(register_json, open(register_json_path, "w"), indent=4)

    print("\n", "-" * 10, "Installation is Registered", "-" * 10)



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from compas_3gs.rhino.artists.volmeshartist import VolMeshArtist


__all__ = ['ForceVolMeshArtist']


class ForceVolMeshArtist(VolMeshArtist):
    """Artist for visualizing force diagrams in the Rhino model space."""

    def __init__(self, force, layer=None):
        super(ForceVolMeshArtist, self).__init__(force, layer=layer)

    @property
    def vertex_xyz(self):
        """dict:
        The view coordinates of the mesh vertices.
        The view coordinates default to the actual mesh coordinates.
        """
        if not self._vertex_xyz:
            self._vertex_xyz = {vertex: self.volmesh.vertex_attributes(vertex, 'xy') + [0.0] for vertex in self.volmesh.vertices()}
        return self._vertex_xyz

    @vertex_xyz.setter
    def vertex_xyz(self, vertex_xyz):
        self._vertex_xyz = vertex_xyz



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from compas.geometry import normalize_vector
from compas.geometry import scale_vector
from compas.geometry import add_vectors
from compas.geometry import dot_vectors
from compas.geometry import angle_vectors
from compas.geometry import length_vector
from compas.geometry import centroid_points

from compas_3gs.algorithms import volmesh_planarise

from compas_3gs.utilities import print_result


__all__ = ['volmesh_reciprocate']


def volmesh_reciprocate(volmesh,
                        formdiagram,

                        kmax=100,
                        weight=1.0,

                        fix_vkeys=[],

                        edge_min=0.1,
                        edge_max=10000,

                        tolerance=1.0,
                        tolerance_boundary=1.0,

                        callback=None,
                        callback_args=None,

                        print_result_info=False):
    """Perpendicularizes the faces of the polyhedral force diagram to the corresponding dual edges in the polyhedral form diagram.

    Parameters
    ----------
    volmesh : VolMesh
        A volmesh object representing a polyhedral force diagram.
    formdiagram : VolMesh or Network
        A network object representing a polyhedral form diagram.
    kmax : int, optional [100]
        Maximum number of iterations.
    weight : float, optional [1.0]
        A float, between 0 and 1, which determines how much each diagram changes. 1 changes the form diagram only, while 0 changes the force diagram only. Default is ``1.0``.
    fix_vkeys : list, optional []
        List of vkeys to fix.
    edge_min : float, optional [None]
        Value for minimum edge length to be imposed.
    edge_max : float, optional [None]
        Value for maximum edge length allowed.
    tolerance: float, optional [0.001]
        Value for convergence tolerance. Deviation is measured by the the dot product of the input and target vectors.
    callback : callable, optional [None]
        A user-defined callback function to be executed after every iteration.
    callback_args : tuple, optional [None]
        Additional parameters to be passed to the callback.
    print_result_info : bool, optional
        If True, print the result of the algorithm.

    Raises
    ------
    Exception
        If a callback is provided, but it is not callable.

    Notes
    -----
    The orientations of the boundary faces of the polyhedral force diagram are always fixed by default.

    .. seealso ::
        compas.geometry.network_parallelise_edges

    """

    if callback:
        if not callable(callback):
            raise Exception('Callback is not callable.')

    free_vkeys = list(set(formdiagram.vertices()) - set(fix_vkeys))

    init_normals = {fkey: volmesh.halfface_normal(fkey) for fkey in volmesh.faces()}

    boundary_fkeys = volmesh.halffaces_on_boundaries()

    # --------------------------------------------------------------------------
    #   1. compute target vectors
    # --------------------------------------------------------------------------
    target_vectors = {}
    target_normals = {}

    for u, v in formdiagram.edges():
        u_hfkey = volmesh.cell_pair_halffaces(u, v)[0]
        face_normal = scale_vector(volmesh.halfface_normal(u_hfkey), weight)
        edge_vector = scale_vector(formdiagram.edge_vector(u, v), 1 - weight)
        target = normalize_vector(add_vectors(face_normal, edge_vector))
        target_vectors[(u, v)] = {'fkey': u_hfkey,
                                  'target': target}
        target_normals[u_hfkey] = target

    for fkey in boundary_fkeys:
        target_normals[fkey] = init_normals[fkey]

    # --------------------------------------------------------------------------
    #   2. loop
    # --------------------------------------------------------------------------
    for k in range(kmax):

        deviation_boundary_perp = 0

        # ----------------------------------------------------------------------
        #   3. update form diagram
        # ----------------------------------------------------------------------
        if weight != 0:

            new_form_xyz = {vkey: [] for vkey in formdiagram.vertices()}

            for u, v in target_vectors:
                target_v = target_vectors[(u, v)]['target']
                edge_v = formdiagram.edge_vector(u, v, unitized=False)

                # target edge length -------------------------------------------
                l = length_vector(edge_v)

                l_min = formdiagram.edge_attribute((u, v), 'l_min')
                l_max = formdiagram.edge_attribute((u, v), 'l_max')

                # min edge
                if edge_min:
                    l_min = edge_min
                if l < l_min:
                    l = l_min

                # max edge
                if edge_max:
                    l_max = edge_max
                if l > l_max:
                    l = l_max

                # check edge orientation ---------------------------------------
                direction = _get_lambda(edge_v, target_v)
                l *= direction

                # collect new coordinates --------------------------------------
                if u in free_vkeys:
                    new_u_xyz = add_vectors(formdiagram.vertex_coordinates(v), scale_vector(target_v, -1 * l))
                    new_form_xyz[u].append(new_u_xyz)

                if v in free_vkeys:
                    new_v_xyz = add_vectors(formdiagram.vertex_coordinates(u), scale_vector(target_v, l))
                    new_form_xyz[v].append(new_v_xyz)

            # compute new vertex coordinates -----------------------------------
            for vkey in free_vkeys:
                final_xyz = centroid_points(new_form_xyz[vkey])
                formdiagram.vertex_update_xyz(vkey, final_xyz)

        # ----------------------------------------------------------------------
        #   4. update force diagram
        # ----------------------------------------------------------------------

        if weight != 1:
            volmesh_planarise(volmesh,
                              kmax=1,
                              target_normals=target_normals)

        # boundary perpness
        for fkey in boundary_fkeys:
            f_normal = volmesh.halfface_normal(fkey)
            target_normal = target_normals[fkey]
            b_perpness = angle_vectors(f_normal, target_normal, deg=True)
            # 1 - abs(dot_vectors(f_normal, target_normal))

            if b_perpness > deviation_boundary_perp:
                deviation_boundary_perp = b_perpness

        # ----------------------------------------------------------------------
        #   5. check convergence
        # ----------------------------------------------------------------------
        perpness = _check_deviation(volmesh, formdiagram)

        if perpness < tolerance and deviation_boundary_perp < tolerance_boundary:

            if print_result_info:
                print_result('Reciprocation', k, perpness)

            break

        # callback / conduit ---------------------------------------------------
        if callback:
            callback(volmesh, formdiagram, k, callback_args)


def cellnetwork_reciprocate(cellnetwork):
    raise NotImplementedError


# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   reciprocation helpers
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


def _get_lambda(vector_1, vector_2):
    dot = dot_vectors(vector_1, vector_2)
    if dot < 0:
        return -1
    else:
        return 1


def _check_deviation(volmesh, network):
    deviation = 0
    for u, v in network.edges():
        u_hf, v_hf = volmesh.cell_pair_halffaces(u, v)
        normal = volmesh.halfface_normal(u_hf)
        edge = network.edge_vector(u, v, unitized=True)
        perp_check = angle_vectors(normal, edge, deg=True)
        # dot = dot_vectors(normal, edge)
        # perp_check = 1 - abs(dot)
        if perp_check > deviation:
            deviation = perp_check
    return deviation


# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   Main
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


if __name__ == '__main__':
    pass



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import scriptcontext as sc

import compas_rhino

from compas_rhino.objects.select import mesh_select_face

from compas_3gs.rhino import rhino_volmesh_pull_halffaces

from compas_3gs.rhino import VolmeshHalffaceInspector

__commandname__ = "TGS_force_pull_halffaces"


def RunCommand(is_interactive):

    if '3GS' not in sc.sticky:
        compas_rhino.display_message('3GS has not been initialised yet.')
        return

    scene = sc.sticky['3GS']['scene']

    # get ForceVolMeshObject from scene
    objects = scene.find_by_name('force')
    if not objects:
        compas_rhino.display_message("There is no force diagram in the scene.")
        return
    force = objects[0]

    current_setting = force.settings['show.faces']
    if not current_setting:
        force.settings['show.faces'] = True
        scene.update()

    # select halfface ----------------------------------------------------------
    boundary_halffaces = force.diagram.halffaces_on_boundaries()

    hf_inspector = VolmeshHalffaceInspector(
        force.diagram,
        hfkeys=boundary_halffaces,
        dependents=True)
    hf_inspector.enable()
    halfface = mesh_select_face(force.diagram)
    hf_inspector.disable()
    del hf_inspector

    # pull faces ---------------------------------------------------------------
    rhino_volmesh_pull_halffaces(force.diagram, hfkey=halfface)

    force.settings['show.faces'] = current_setting

    # check if there is form diagram -------------------------------------------
    objects = scene.find_by_name('form')
    if not objects:
        force.check_eq()
        scene.update()
        return
    form = objects[0]

    # update -------------------------------------------------------------------
    form.diagram.update_angle_deviations()

    form.check_eq()
    force.check_eq()

    scene.update()
    scene.save()


# ==============================================================================
# Main
# ==============================================================================

if __name__ == "__main__":

    RunCommand(True)



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division


from .formnetworkartist import *  # noqa: F401 F403
from .forcevolmeshartist import *  # noqa: F401 F403
from .networkartist import *  # noqa: F401 F403
from .volmeshartist import *  # noqa: F401 F403

from compas_3gs.diagrams import ForceVolMesh
from compas_3gs.diagrams import FormNetwork

from .forcevolmeshartist import ForceVolMeshArtist
from .formnetworkartist import FormNetworkArtist
from .volmeshartist import VolMeshArtist
from .networkartist import NetworkArtist

VolMeshArtist.register(ForceVolMesh, ForceVolMeshArtist)
NetworkArtist.register(FormNetwork, FormNetworkArtist)


__all__ = [name for name in dir() if not name.startswith('_')]



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division


__all__ = ['pair_hf_to_uv',
           'pair_uv_to_hf']


def pair_hf_to_uv(volmesh, network):
    """Pairs the directed halffaces of a volmesh to the corresponding edge (u, v) of the dual network.

    Parameters
    ----------
    volmesh : VolMesh
        A volmesh object representing a polyhedral force diagram.
    network : Network
        A network object representing a polyhedral form diagram.

    Returns
    -------
    dictionary
        A dictionary of u_hfkey-(u, v) pairs.

    Notes
    -----
    u_hfkey is an interior halfface of the volmesh that belongs to volmesh.cell[u], which points to volmesh.cell[v].
    In another words, its pair (or opposite) halfface belongs to volmesh.cell[v].

    """
    hf_uv_dict = {}

    for u, v in network.edges():
        u_hfkey, v_hfkey = volmesh.cell_pair_halffaces(u, v)
        hf_uv_dict[u_hfkey] = (u, v)
        hf_uv_dict[v_hfkey] = (v, u)

    return hf_uv_dict


def pair_uv_to_hf(network, volmesh):
    """Pairs the directed edges (u, v) of a network to the corresponding halffaces of the dual volmesh.

    Parameters
    ----------
    volmesh : VolMesh
        A volmesh object representing a polyhedral force diagram.
    network : Network
        A network object representing a polyhedral form diagram.

    Returns
    -------
    dictionary
        A dictionary of (u, v)-u_hfkey pairs.

    """
    uv_hf_dict = {}

    for u, v in network.edges():
        u_hfkey, v_hfkey = volmesh.cell_pair_halffaces(u, v)
        uv_hf_dict[(u, v)] = u_hfkey

    return uv_hf_dict


# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   Main
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


if __name__ == '__main__':
    pass
